#include <avr/io.h>
#include <avr/interrupt.h>

#include "lights.h"

#define AMBIENT_ADC N_PWMLEDS
#define BATTERY_ADC (N_PWMLEDS + 1)

#define NUM_ADCS 5
volatile static unsigned char current_adc;
static uint16_t adc_sum;
static unsigned char sum_shift;
static unsigned char adc_vals;

static void inline setup_mux(unsigned char n)
{
	ADCSRA |= _BV(ADEN); // enable ADC

	/* ADC numbering: PWM LEDs first, then ambient light sensor, battery sensor */
	switch (n) {
	case 0: // pwmled 1: 1.1V, ADC0,1 (PA0,1), gain 20
		ADMUX = _BV(REFS1) | _BV(MUX3) | _BV(MUX1) | _BV(MUX0);
		sum_shift = 3; // 8 measurements
		break;
	case 1: // pwmled 2: 1.1V, ADC2,1 (PA2,1), gain 20
		ADMUX = _BV(REFS1) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
		sum_shift = 3; // 8 measurements
		break;
	case 2: // pwmled 3: 1.1V, ADC4 (PA5), single-ended
		ADMUX = _BV(REFS1) | _BV(MUX2);
		sum_shift = 3; // 8 measurements
		break;
	case 3: // ambient light: 1.1V, ADC5 (PA6), single-ended
		ADMUX = _BV(REFS1) | _BV(MUX2) | _BV(MUX0);
		sum_shift = 0; // 1 measurement
		break;
	case 4: // batt voltage: 1.1V, ADC6 (PA7), single-ended
		ADMUX = _BV(REFS1) | _BV(MUX2) | _BV(MUX1);
		sum_shift = 0; // 1 measurement
		break;
	}

	adc_sum = 0;
	adc_vals = 1 << sum_shift;
}

static void start_next_adc()
{
	if (current_adc > 0) {
		current_adc--;
		// set up mux, start one-shot conversion
		setup_mux(current_adc);
		ADCSRA |= _BV(ADSC);
	} else {
		current_adc = NUM_ADCS;
		// TODO: kick the watchdog here.
	}
}

void init_adc()
{
	current_adc = NUM_ADCS;

	ADCSRA = _BV(ADEN)			// enable
		| _BV(ADPS1) | _BV(ADPS0)	// CLK/8 = 125 kHz
		// | _BV(ADPS2)			// CLK/16 = 62.5 kHz
		;
	// ADCSRB |= _BV(GSEL); // gain 8 or 32

	// Disable digital input on all bits used by ADC
	DIDR0 = _BV(ADC0D) | _BV(ADC1D) | _BV(ADC2D)
		| _BV(ADC4D) | _BV(ADC5D) | _BV(ADC6D);

	ADCSRA |= _BV(ADSC);

	/* Do first conversion and drop the result */
	while ((ADCSRA & _BV(ADIF)) == 0)
		;
	ADCSRA |= _BV(ADIF); // clear the IRQ flag
	ADCSRA |= _BV(ADIE); // enable IRQ

	ADCSRA &= ~_BV(ADEN); // disable until needed
}

void susp_adc()
{
	ADCSRA = 0;
	DIDR0 = 0;
}

ISR(ADC_vect) { // IRQ handler
	uint16_t adcval = ADCW;

	if (adc_vals)
		// start the next conversion immediately
		ADCSRA |= _BV(ADSC);
	else
		ADCSRA &= ~_BV(ADEN); // the last one, disable ADC

	if (adc_vals < (1 << sum_shift))
		 // drop the first conversion, use all others
		 adc_sum += adcval;

	if (adc_vals) {
		adc_vals--;
		return;
	}

	// Now handle the (1 << sum_shift) measurements

	adcval = adc_sum >> sum_shift;

	if (current_adc < N_PWMLEDS)
		pwmled_adc(current_adc, adcval);
	if (current_adc == AMBIENT_ADC)
		ambient_adc(adcval);
	if (current_adc == BATTERY_ADC)
		battery_adc(adcval);
	
	start_next_adc();
}

void timer_start_adcs()
{
	if (current_adc == NUM_ADCS) // Don't start if in progress
		start_next_adc();
}

