#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>

#include "lights.h"

static uint16_t button_start[N_BUTTONS];
static unsigned char prev_pin;

static unsigned char user_params[MAX_USER_PARAMS] = { 0, 0, 0 };
static unsigned char user_params_max[MAX_USER_PARAMS] = { 3, 2, 2 };

static unsigned char user_params_state = 0;
	/*
	 * Here 0 means "no setup currently in progress",
	 * 1 .. MAX_USER_PARAMS means "now short presses increase or decrease
	 *           the value of user_params[user_params_state-1]"
	 */

static uint16_t user_params_starttime;

static void inline set_status_led(unsigned char n, pattern_t *pattern)
{
	led_set_pattern(n + N_PWMLEDS, pattern);
}

pattern_t *status_pattern_select(unsigned char n)
{
	if (user_params_state) // Setup in progress
		return number_pattern(n == 0
			? user_params_state
			: 1 + user_params[user_params_state-1]
		);
	else if (!(prev_pin & _BV(PA3)) || !(prev_pin & _BV(PA4)))
		// at least one button is pressed
		return off_pattern;
	// otherwise, an ordinary state
	if (n == 0) {
		// error led
		return number_pattern(1 + ambient_zone);
	} else {
		return off_pattern; // for now
	}
}

static void toggle_bright_mode()
{
	// TODO
}

static void set_panic_mode()
{
	// TODO
}

unsigned char get_user_param(unsigned char param)
{
	if (param < MAX_USER_PARAMS)
		return user_params[param];
	return 0; // FIXME: internal error?
}

static inline void short_press(unsigned char button)
{
	unsigned char param;

	if (user_params_state == 0) {
		if (button == 0)
			toggle_bright_mode();
		else
			set_panic_mode();
		return;
	}

	param = user_params_state-1;

	if (button == 0) {
		if (user_params[param])
			user_params[param]--;
		else
			user_params[param] = user_params_max[param];
	} else {
		user_params[param]++;

		if (user_params[param] >= user_params_max[param])
			user_params[param] = 0;
	}
	// FIXME: notify somebody about user_params change?

	set_status_led(1, status_pattern_select(1));

	user_params_starttime = jiffies;
}

static inline void long_press(unsigned char button)
{
	if (button == 0) {
		power_down();
		return;
	}

	// button 1 - cycle through states
	user_params_state++;

	if (user_params_state >= MAX_USER_PARAMS)
		user_params_state = 1;

	set_status_led(0, status_pattern_select(0));
	set_status_led(1, status_pattern_select(1));
	user_params_starttime = jiffies;
}

void init_buttons()
{
	DDRA &= ~(_BV(PA3) | _BV(PA4));
	PORTA |=  _BV(PA3) | _BV(PA4);

	button_start[0] = 0;
	button_start[1] = 0;
	prev_pin = _BV(PA3) | _BV(PA4);
	user_params_state = 0;
}

static void handle_button(unsigned char button, unsigned char cur,
	unsigned char prev)
{
	// BEWARE: pins are at _zero_ when pressed!
	if (!cur && prev) {                   // --- just pressed ---
		button_start[button] = jiffies;
		set_status_led(button, off_pattern);

	} else if (!cur && !prev) {           // --- is still pressed ---
		uint16_t duration = jiffies - button_start[button];

		if (duration > 80) {
			set_status_led(button, on_pattern);
				// acknowledge long press
		}
	} else if (cur && !prev) {            // --- just released ---
		uint16_t duration = jiffies - button_start[button];

		if (duration > 6 && duration < 30) {
			short_press(button);
		} else if (duration > 80) {
			set_status_led(button, off_pattern);
			long_press(button);
		}
		// ignore other button-press durations
	}
}

void timer_check_buttons()
{
	unsigned char pin = PINA & (_BV(PA3) | _BV(PA4));

	handle_button(0, pin & _BV(PA3), prev_pin & _BV(PA3));
	handle_button(1, pin & _BV(PA4), prev_pin & _BV(PA4));

	prev_pin = pin;

	if (jiffies - user_params_starttime > 500) {
		user_params_state = 0;
		set_status_led(0, status_pattern_select(0));
		set_status_led(1, status_pattern_select(1));
	}
}

